from typing import Dict, List
from astroid import nodes
from pylint.checkers import BaseChecker
from pylint.checkers import utils as checker_utils
from pylint.interfaces import IAstroidChecker


class ListChecker(BaseChecker):
    """
    Check for inefficient list usage
    """

    __implements__ = IAstroidChecker

    name = 'list-checker'
    priority = -1
    msgs = {
        'W8301': (
            'Use tuple instead of list for a non-mutated sequence',
            'use-tuple-over-list',
            ''
        ),
    }

    def __init__(self, linter=None):
        super().__init__(linter)
        self._lists_to_watch: List[Dict[str, nodes.AssignName]] = []

    def visit_assign(self, node: nodes.Assign):
        if not isinstance(node.value, nodes.List):
            return
        if len(node.targets) > 1:
            return
        if isinstance(node.targets[0], nodes.AssignName):
            self._lists_to_watch[-1][node.targets[0].name] = node.targets[0]

    def visit_module(self, node: nodes.Module):
        self._lists_to_watch.append({})

    def _raise_for_scope(self):
        _lists = self._lists_to_watch.pop()
        for _assignment in _lists.values():
            self.add_message("use-tuple-over-list", node=_assignment.parent.value)

    @checker_utils.check_messages("use-tuple-over-list")
    def leave_module(self, node: nodes.Module):
        self._raise_for_scope()

    def visit_functiondef(self, node: nodes.FunctionDef):
        self._lists_to_watch.append({})

    @checker_utils.check_messages("use-tuple-over-list")
    def leave_functiondef(self, node: nodes.FunctionDef):
        self._raise_for_scope()

    def visit_call(self, node: nodes.Call) -> None:
        """Look for method calls to list nodes."""
        if not isinstance(node.func, nodes.Attribute):
            return
        if not isinstance(node.func.expr, nodes.Name):
            return
        # TODO : Filter from non-mutation methods
        self._mark_mutated(node.func.expr)

    def _mark_mutated(self, _name: nodes.Name):
        if _name.name in self._lists_to_watch[-1]:
            del self._lists_to_watch[-1][_name.name]
            return
        scope, _ = _name.lookup(_name.name)
        if not isinstance(scope, nodes.Module):
            return
        if _name.name in scope.globals \
            and _name.name in self._lists_to_watch[0]:
            del self._lists_to_watch[0][_name.name]

    def visit_subscript(self, node: nodes.Subscript):
        if not isinstance(node.parent, nodes.Assign):
            return
        if not isinstance(node.value, nodes.Name):
            return
        self._mark_mutated(node.value)



MY_GLOBAL_CONSTANT_C = 1234
MY_GLOBAL_CONSTANT_A = 3.14
MY_GLOBAL_VARIABLE = 1234

def global_constant_in_loop():
    """Do a quick sum."""

    total = MY_GLOBAL_CONSTANT_A
    for i in range(10_000):
        total += i * MY_GLOBAL_CONSTANT_C # [loop-invariant-global-usage]


def global_variable_in_loop():
    """Use the module global as a global."""
    global MY_GLOBAL_VARIABLE # [global-statement]

    total = MY_GLOBAL_CONSTANT_A
    for i in range(10_000):
        MY_GLOBAL_VARIABLE += total
        total += i

def recursive_example(x):
    for i in range(100):
        recursive_example(i)

    for i in range(100):
        global_constant_in_loop()



import astroid
import perflint.for_loop_checker
from base import BaseCheckerTestCase


class TestUniqueReturnChecker(BaseCheckerTestCase):
    CHECKER_CLASS = perflint.for_loop_checker.LoopInvariantChecker

    def test_basic_loop_invariant(self):
        test_node = astroid.extract_node("""
        def test(): #@
            items = (1,2,3,4)

            for item in list(items):
                x = print("There are ", len(items), "items")
        """)

        with self.assertAddedMessage("loop-invariant-statement"):
            self.walk(test_node)

    def test_basic_loop_invariant_while(self):
        test_node = astroid.extract_node("""
        def test(): #@
            items = (1,2,3,4)
            i = 0
            while i < len(items):
                x = print("There are ", len(items), "items")
                i += 1
        """)

        with self.assertAddedMessage("loop-invariant-statement"):
            self.walk(test_node)

    def test_basic_loop_variant_while(self):
        test_node = astroid.extract_node("""
        def test(): #@
            items = (1,2,3,4)
            i = 0
            while i < len(items):
                i += 1
                print(i)
        """)

        with self.assertAddedMessage("loop-invariant-statement"):
            self.walk(test_node)

    def test_kwarg_usage_in_while(self):
        test_node = astroid.extract_node("""
        def foo(arg):
            pass

        def test(): #@
            items = (1,2,3,4)
            i = 0
            while i < len(items):
                foo(arg=i)
        """)

        with self.assertNoMessages():
            self.walk(test_node)

    def test_basic_loop_variant_by_method(self):
        test_node = astroid.extract_node("""
        def test(): #@
            items = [1,2,3,4]

            for item in items:
                x = print("There are ", len(items), "items")
                items.clear()
        """)

        with self.assertNoMessages():
            self.walk(test_node)

    def test_global_in_for_loop(self):
        test_func = astroid.extract_node("""
        glbl = 1

        def test(): #@
            items = (1,2,3,4)

            for item in list(items):
                glbl
        """)

        with self.assertAddedMessage("loop-invariant-global-usage"):
            self.walk(test_func)

    def test_assigned_global_in_for_loop(self):
        test_func = astroid.extract_node("""
        glbl = 1

        def test(): #@
            items = (1,2,3,4)

            for glbl in list(items):
                glbl
        """)

        with self.assertNoMessages():
            self.walk(test_func)

    def test_self_assignment_for_loop(self):
        test_func = astroid.extract_node("""
        class Foo:
            n = 1

            def loop(self):
                for self.n in range(4):
                    print(self.n)
        
        def test(): #@
            f = Foo()
            f.loop()
        """)

        with self.assertNoMessages():
            self.walk(test_func)

    def test_byte_slice(self):
        test_func = astroid.extract_node("""
        def test(): #@
            word = b'word'

            for i in range(10):
                word[0:i]
        """)

        with self.assertAddedMessage("memoryview-over-bytes"):
            self.walk(test_func)
    
    def test_byte_slice_as_arg(self):
        test_func = astroid.extract_node("""
        def test(arg1: bytes): #@
            for i in range(10):
                arg1[0:i]
        """)

        with self.assertAddedMessage("memoryview-over-bytes"):
            self.walk(test_func)
        
    def test_dotted_name_in_loop(self):
        test_func = astroid.extract_node("""
        import os
        def test(): #@
            for item in items:
                os.environ[item]
        """)

        with self.assertAddedMessage("dotted-import-in-loop"):
            self.walk(test_func)

    def test_worse_dotted_name_in_loop(self):
        test_func = astroid.extract_node("""
        import os
        def test(): #@
            for item in items:
                os.path.exists(item)
        """)

        with self.assertAddedMessage("dotted-import-in-loop"):
            self.walk(test_func)

    def test_variant_f_string_in_loop(self):
        test_func = astroid.extract_node("""
        def test(): #@
            n = 0
            for i in range(2):
                print(f"{n} {i}")
        """)

        with self.assertNoMessages():
            self.walk(test_func)

    def test_invariant_f_string_in_loop(self):
        test_func = astroid.extract_node("""
        def test(): #@
            i = 2
            for _ in range(2):
                print(f"{i}")
        """)

        with self.assertAddedMessage("loop-invariant-statement"):
            self.walk(test_func)

    def test_call_side_effect(self):
        test_func = astroid.extract_node("""
        class Foo:
            pass

        def test(): #@
            for n in range(1):
                Foo()
        """)

        with self.assertNoMessages():
            self.walk(test_func)
    
    def test_assign_const(self):
        test_func = astroid.extract_node("""
        def test(): #@
            for n in range(1):
                x = None
        """)

        with self.assertNoMessages():
            self.walk(test_func)
    
    def test_slice_fragment(self):
        test_func = astroid.extract_node("""
        def test(): #@
            for n in range(1):
                x = None
        """)

        with self.assertNoMessages():
            self.walk(test_func)
    
    def test_index_fragment(self):
        test_func = astroid.extract_node("""
        def test(): #@
            fruits = ["apple", "banana", "pear"]
            for fruit in fruits:
                print(fruit)
                _ = fruit[1:]
                _ = fruit[-1]
                _ = fruit[::-1] 
        """)

        with self.assertNoMessages():
            self.walk(test_func)
    

    def test_return(self):
        test_func = astroid.extract_node("""
        def test(): #@
            fruits = ["apple", "banana", "pear"]
            for _ in fruits:
                return fruits
        """)

        with self.assertNoMessages():
            self.walk(test_func)
    


from astroid import nodes
from pylint.checkers import BaseChecker
from pylint.checkers import utils as checker_utils
from pylint.interfaces import IAstroidChecker
from astroid.helpers import safe_infer


class ComprehensionChecker(BaseChecker):
    """
    Check for comprehension usage
    """

    __implements__ = IAstroidChecker

    name = "comprehension-checker"
    priority = -1
    msgs = {
        "W8401": (
            "Use a list comprehension instead of a for-loop",
            "use-list-comprehension",
            "",
        ),
        "W8402": (
            "Use a list copy instead of a for-loop",
            "use-list-copy",
            "",
        ),
        "W8403": (
            "Use a dictionary comprehension instead of a for-loop",
            "use-dict-comprehension",
            "",
        ),
    }

    def visit_for(self, node: nodes.For):
        pass

    @checker_utils.check_messages(
        "use-list-comprehension", "use-dict-comprehension", "use-list-copy"
    )
    def leave_for(self, node: nodes.For):
        if len(node.body) != 1:
            return
        if isinstance(node.body[0], nodes.If) and not node.body[0].orelse:
            # TODO : Support a simple, single else statement
            if isinstance(node.body[0].body[0], nodes.Expr):
                if not isinstance(node.body[0].body[0].value, nodes.Call):
                    return
                # Is append call.
                if not isinstance(node.body[0].body[0].value.func, nodes.Attribute):
                    return
                if not node.body[0].body[0].value.func.attrname in ["append", "insert"]:
                    return
                self.add_message("use-list-comprehension", node=node)
            elif isinstance(node.body[0].body[0], nodes.Assign):
                if len(node.body[0].body[0].targets) != 1:
                    return
                if not isinstance(node.body[0].body[0].targets[0], nodes.Subscript):
                    return
                if not isinstance(node.body[0].body[0].targets[0].value, nodes.Name):
                    return
                inferred_value = safe_infer(node.body[0].body[0].targets[0].value)
                if isinstance(inferred_value, nodes.Dict):
                    self.add_message("use-dict-comprehension", node=node)
        elif isinstance(node.body[0], nodes.Expr):
            if not isinstance(node.body[0].value, nodes.Call):
                return
            # Is append call.
            if not isinstance(node.body[0].value.func, nodes.Attribute):
                return
            if not node.body[0].value.func.attrname in ["append", "insert"]:
                return
            self.add_message("use-list-copy", node=node)
        elif isinstance(node.body[0], nodes.Assign):
            if len(node.body[0].targets) != 1:
                return
            if not isinstance(node.body[0].targets[0], nodes.Subscript):
                return
            if not isinstance(node.body[0].targets[0].value, nodes.Name):
                return
            inferred_value = safe_infer(node.body[0].targets[0].value)
            if isinstance(inferred_value, nodes.Dict):
                self.add_message("use-dict-comprehension", node=node)



"""Test ComprehensionChecker"""


def should_be_a_list_copy():
    """Using the copy() method would be more efficient."""
    original = range(10_000)
    filtered = []
    for i in original:
        filtered.append(i)


def should_be_a_list_comprehension_filtered():
    """A List comprehension would be more efficient."""
    original = range(10_000)
    filtered = []
    for i in original:
        if i % 2:
            filtered.append(i)


def should_be_a_dict_comprehension():
    pairs = (("a", 1), ("b", 2))
    result = {}
    for x, y in pairs:
        result[x] = y


def should_be_a_dict_comprehension_filtered():
    pairs = (("a", 1), ("b", 2))
    result = {}
    for x, y in pairs:
        if y % 2:
            result[x] = y


def should_not_be_a_list_comprehension(args):
    """Internal helper for get_args."""
    res = []
    for arg in args:
        if not isinstance(arg, tuple):
            res.append(arg)
        elif is_callable_type(arg[0]):
            if len(arg) == 2:
                res.append(Callable[[], arg[1]])
            elif arg[1] is Ellipsis:
                res.append(Callable[..., arg[2]])
            else:
                res.append(Callable[list(arg[1:-1]), arg[-1]])
        else:
            res.append(type(arg[0]).__getitem__(arg[0], _eval_args(arg[1:])))
    return tuple(res)



import astroid
import perflint.for_loop_checker

from base import BaseCheckerTestCase


class TestUniqueReturnChecker(BaseCheckerTestCase):
    CHECKER_CLASS = perflint.for_loop_checker.ForLoopChecker

    def test_bad_list_cast(self):
        for_node = astroid.extract_node("""
        def test():
            items = (1,2,3,4)

            for item in list(items): #@
                pass
        """)

        with self.assertAddedMessage("unnecessary-list-cast"):
            self.checker.visit_for(for_node)

    def test_bad_list_cast_typed(self):
        for_node = astroid.extract_node("""
        def test(items: List[int]):
            for item in list(items): #@
                pass
        """)

        with self.assertAddedMessage("unnecessary-list-cast"):
            self.checker.visit_for(for_node)

    def test_bad_dict_usage_values(self):
        for_node = astroid.extract_node("""
        def test():
            d = {1: 1, 2: 2}

            for _, v in d.items(): #@
                pass
        """)

        with self.assertAddedMessage("incorrect-dictionary-iterator"):
            self.checker.visit_for(for_node)

    def test_bad_dict_usage_keys(self):
        for_node = astroid.extract_node("""
        def test():
            d = {1: 1, 2: 2}

            for k, _ in d.items(): #@
                pass
        """)

        with self.assertAddedMessage("incorrect-dictionary-iterator"):
            self.checker.visit_for(for_node)



def example_bytes_slice():
    word = b'the lazy brown dog jumped'
    for i in range(10):
        # Memoryview slicing is 10x faster than bytes slicing
        if word[0:i] == 'the':
            return True

def example_bytes_slice_as_arg(word: bytes):
    for i in range(10):
        # Memoryview slicing is 10x faster than bytes slicing
        if word[0:i] == 'the':
            return True



"""Pylint extension with performance anti-patterns"""
from typing import TYPE_CHECKING

from perflint.for_loop_checker import ForLoopChecker, LoopInvariantChecker
from perflint.list_checker import ListChecker
from perflint.comprehension_checker import ComprehensionChecker

if TYPE_CHECKING:
    from pylint.lint import PyLinter

__version__ = "0.7.1"


def register(linter: "PyLinter") -> None:
    """This required method auto registers the checker during initialization.

    :param linter: The linter to register the checker to.
    """

    linter.register_checker(ForLoopChecker(linter))
    linter.register_checker(LoopInvariantChecker(linter))
    linter.register_checker(ListChecker(linter))
    linter.register_checker(ComprehensionChecker(linter))



def mutated_list():
    fruit = ["banana", "pear", "orange"]
    fruit.clear()


def non_mutated_list():
    fruit = ["banana", "pear", "orange"]  # [use-tuple-over-list]
    len(fruit)
    for i in fruit:
        print(i)


def index_mutated_list():
    fruit = ["banana", "pear", "orange"]
    fruit[2] = "mandarin"
    len(fruit)
    for i in fruit:
        print(i)


def index_non_mutated_list():
    fruit = ["banana", "pear", "orange"]  # [use-tuple-over-list]
    print(fruit[2])
    len(fruit)
    for i in fruit:
        print(i)


def slice_mutated_list():
    fruit = ["banana", "pear", "orange"]
    fruit[0:1] = ("grapes", "plum")
    len(fruit)
    for i in fruit:
        print(i)



import os


def foo(x):
    pass


def loop_invariant_statement():
    """Catch basic loop-invariant function call."""
    x = (1, 2, 3, 4)

    for i in range(10_000):
        # x is never changed in this loop scope,
        # so this expression should be evaluated outside
        print(len(x) * i)  # [loop-invariant-statement]


def loop_invariant_statement_more_complex():
    """Catch basic loop-invariant function call."""
    x = [1, 2, 3, 4]
    i = 6

    for j in range(10_000):
        # x is never changed in this loop scope,
        # so this expression should be evaluated outside
        print(len(x) * i + j)  # [loop-invariant-statement]


def loop_invariant_statement_method_side_effect():
    """Catch basic loop-invariant function call."""
    x = [1, 2, 3, 4]
    i = 6

    for j in range(10_000):
        print(len(x) * i + j)
        x.clear()  # x changes as a side-effect


def loop_invariant_branching():
    """Ensure node is walked up to find a loop-invariant branch"""
    x = [1, 2, 3, 4]
    i = 6

    for j in range(10_000):
        # Marks entire branch
        if len(x) > 2:  # [loop-invariant-statement]
            print(x * i)  # [loop-invariant-statement]

    # Marks comparator, but not print
    for j in range(10_000):
        if len(x) > 2:  # [loop-invariant-statement]
            print(x * j)


def loop_invariant_statement_side_effect_function():
    """Catch basic loop-invariant function call."""
    x = [1, 2, 3, 4]
    i = 6
    _len = len

    def len(x):  # now with side effects!
        x.clear()
        return 0

    for j in range(10_000):
        # x is never changed in this loop scope,
        # so this expression should be evaluated outside
        print(len(x) + j)

    len = _len


def loop_invariant_statement_but_name():
    """Catch basic loop-invariant function call."""
    i = 6

    for _ in range(10_000):
        i


def loop_invariant_statement_while():
    """Catch basic loop-invariant function call."""
    x = (1, 2, 3, 4)
    i = 0
    while i < 100:
        i += 1
        # x is never changed in this loop scope,
        # so this expression should be evaluated outside
        print(len(x) * i)  # [loop-invariant-statement]
        y = x[0] + x[1]  # [loop-invariant-statement]
        foo(x=x)


def loop_invariant_statement_more_complex_while():
    """Catch basic loop-invariant function call."""
    x = [1, 2, 3, 4]
    i = 6
    j = 0
    while j < 100:
        # x is never changed in this loop scope,
        # so this expression should be evaluated outside
        print(len(x) * i + j)  # [loop-invariant-statement]


def loop_invariant_statement_method_side_effect_while():
    """Catch basic loop-invariant function call."""
    x = [1, 2, 3, 4]
    i = 6
    j = 0
    while j < 10_000:
        j += 1
        print(len(x) * i + j)
        x.clear()  # x changes as a side-effect


def loop_invariant_branching_while():
    """Ensure node is walked up to find a loop-invariant branch"""
    x = [1, 2, 3, 4]
    i = 6
    j = 0
    while j < 10_000:
        j += 1
        # Marks entire branch
        if len(x) > 2:  # [loop-invariant-statement]
            print(x * i)  # [loop-invariant-statement]

    # Marks comparator, but not print
    j = 0
    while j < 10_000:
        j += 1
        if len(x) > 2:  # [loop-invariant-statement]
            print(x * j)


def loop_invariant_statement_side_effect_function_while():
    """Catch basic loop-invariant function call."""
    x = [1, 2, 3, 4]
    i = 6
    _len = len

    def len(x):  # now with side effects!
        x.clear()
        return 0

    j = 0
    while j < 10_000:
        j += 1
        # x is never changed in this loop scope,
        # so this expression should be evaluated outside
        print(len(x) + j)  # [loop-invariant-statement]

    len = _len


def loop_invariant_statement_but_name_while():
    """Catch basic loop-invariant function call."""
    i = 6

    for _ in range(10_000):
        i  # [loop-invariant-statement]


def test_dotted_import(items):
    for item in items:
        val = os.environ[item]


def even_worse_dotted_import(items):
    for item in items:
        val = os.path.exists(item)


def loop_invariance_in_self_assignment():
    class Foo:
        n = 1

        def loop(self):
            i = 4
            for self.n in range(4):
                print(self.n)
                len(i)  # [loop-invariant-statement]

    def test():  # @
        f = Foo()
        f.loop()

    test()


def invariant_fstrings():
    i = 1
    for n in range(2):
        print(f"{i}")  # [loop-invariant-statement]
        print(f"{n}")
        print(f"{i} + {n}")
        print(f"{n} + {i}")


def invariant_literals():
    i = 1
    for n in range(2):
        d = {"x": i}  # [loop-invariant-statement]
        d2 = {"x": i, "j": n}
        d3 = {"j": n}

    for n in range(2):
        print("x" * n)

    for n in range(2):
        print("x" * i)  # [loop-invariant-statement]


def invariant_iteration_sub():  # @
    items = (1, 2, 3, 4)

    for _ in items:
        x = print("There are ", len(items), "items")  # [loop-invariant-statement]


def invariant_consts():
    for _ in range(4):
        len("BANANANANANAN")  # [loop-invariant-statement]
        len((1, 2, 3, 4))  # [loop-invariant-statement]
        max((1, 2, 3, 4))  # [loop-invariant-statement]
        type(None)  # [loop-invariant-statement]


def invariant_slices():  # @
    l = (1, 2, 3, 4)
    for n in range(1):
        _ = l[1:2]  # [loop-invariant-statement]
        _ = l[n:3]
        _ = l[1]  # [loop-invariant-statement]


def variant_slices():  # @
    fruits = ["apple", "banana", "pear"]
    for fruit in fruits:
        print(fruit)
        _ = fruit[1:]
        _ = fruit[-1]
        _ = fruit[::-1]




def simple_static_list():
    """Test warning for casting a list to a list."""
    items = [1, 2, 3]
    for i in list(items): # [unnecessary-list-cast]
        print(i)

def simple_static_tuple():
    """Test warning for casting a tuple to a list."""
    items = (1, 2, 3)
    for i in list(items): # [unnecessary-list-cast]
        print(i)

def simple_static_set():
    """Test warning for casting a set to a list."""
    items = {1, 2, 3}
    for i in list(items): # [unnecessary-list-cast]
        print(i)

def simple_dict_keys():
    """Check that dictionary .items() is being used correctly. """
    fruit = {
        'a': 'Apple',
        'b': 'Banana',
    }

    for _, value in fruit.items(): # [incorrect-dictionary-iterator]
        print(value)

    for key, _ in fruit.items(): # [incorrect-dictionary-iterator]
        print(key)



import contextlib
import pylint.testutils


class BaseCheckerTestCase(pylint.testutils.CheckerTestCase):
    """Extends the basic pylint test case with easier helpers."""

    @contextlib.contextmanager
    def assertAddedMessage(self, message):
        """Assert that a msg_id occurred."""
        yield
        got = [msg.msg_id for msg in self.linter.release_messages()]
        got_str = "\n".join(repr(m) for m in got)
        msg = (
            "Expected messages did not match actual.\n"
            f"Got:\n{got_str}\n"
        )
        assert message in got, msg



from typing import Dict, List, Set, Union
from astroid import nodes
from astroid.helpers import safe_infer
from pylint.checkers import BaseChecker
from pylint.checkers import utils as checker_utils
from pylint.interfaces import IAstroidChecker

iterable_types = (nodes.Tuple, nodes.List, nodes.Set, )
iterable_type_names = ("tuple", "list", "set", )

def get_children_recursive(node: nodes.NodeNG):
    """Get children of a node."""
    for child in node.get_children():
        yield child
        yield from get_children_recursive(child)


def local_type(name: nodes.NodeNG) -> Union[None, nodes.Name]:
    if not isinstance(name, nodes.Name):
        return

    if name.name in name.frame().locals:
        vals = name.frame().locals[name.name]
        if len(vals) > 0:
            assigned = vals[0].assign_type()
            if isinstance(assigned, nodes.Arguments):
                for annotation, arg in zip(assigned.annotations, assigned.arguments):
                    if arg.name == name.name:
                        if isinstance(annotation, nodes.Name):
                            return annotation
                        elif isinstance(annotation, nodes.Subscript) and isinstance(annotation.value, nodes.Name):
                            return annotation.value
                        else:
                            return None
    else:
        return

class ForLoopChecker(BaseChecker):
    """
    Check for poor for-loop usage.
    """

    __implements__ = IAstroidChecker

    name = 'for-loop-checker'
    priority = -1
    msgs = {
        'W8101': (
            'Unnecessary using of list() on an already iterable type.',
            'unnecessary-list-cast',
            'Eager iteration of an iterable is inefficient.'
        ),
        'W8102': (
            'Incorrect iterator method for dictionary, use %s.',
            'incorrect-dictionary-iterator',
            'Incorrect use of .items() when not unpacking key and value.'
        )
    }

    @checker_utils.check_messages("unnecessary-list-cast", "incorrect-dictionary-iterator")
    def visit_for(self, node: nodes.For) -> None:
        """Visit for loops."""
        if not node.iter:
            return
        if not isinstance(node.iter, nodes.Call):
            return
        if not node.iter.func:
            return

        # We have multiple types of checkers, function call, or method calls
        if isinstance(node.iter.func, nodes.Name):
            if not node.iter.args:
                return
            if node.iter.func.name != 'list':
                return

            inferred_value = safe_infer(node.iter.args[0])
            if inferred_value:
                if isinstance(inferred_value, iterable_types):
                    self.add_message("unnecessary-list-cast", node=node.iter)
            else:
                loc = local_type(node.iter.args[0])
                if loc and loc.name.lower() in iterable_type_names:
                    self.add_message("unnecessary-list-cast", node=node.iter)

        elif isinstance(node.iter.func, nodes.Attribute):
            if node.iter.args:  # items() never has a list of arguments!
                return
            if node.iter.func.attrname != 'items':
                return
            if not isinstance(node.target, nodes.Tuple):
                return
            if not len(node.target.elts) == 2:
                return
            if isinstance(node.target.elts[0], nodes.AssignName) and node.target.elts[0].name == '_':
                self.add_message("incorrect-dictionary-iterator", node=node.iter, args=('values()', ))
            if isinstance(node.target.elts[1], nodes.AssignName) and node.target.elts[1].name == '_':
                self.add_message("incorrect-dictionary-iterator", node=node.iter, args=('keys()', ))

        else:
            return



class LoopInvariantChecker(BaseChecker):
    """
    Check for poor for-loop usage.
    """

    __implements__ = IAstroidChecker

    name = 'loop-invariant-checker'
    priority = -1
    msgs = {
        'W8201': (
            'Consider moving this expression outside of the loop.',
            'loop-invariant-statement',
            'None of the variables referred to in this expression change within the loop.'
        ),
        'W8202': (
            'Lookups of global names within a loop is inefficient, copy to a local variable outside of the loop first.',
            'loop-invariant-global-usage',
            'Global name lookups in Python are slower than local names.'
        ),
        'R8203': (
            'Try..except blocks have an overhead. Avoid using them inside a loop unless you\'re using them for control-flow.',
            'loop-try-except-usage',
            'Avoid using try..except within a loop.'
        ),
        'W8204': (
            'Looped slicing of bytes objects is inefficient. Use a memoryview() instead',
            'memoryview-over-bytes',
            'Avoid using byte slicing in loops.'
        ),
        'W8205': (
            'Importing the "%s" name directly is more efficient in this loop.',
            'dotted-import-in-loop',
            'Dotted global names in loops are inefficient.'
        ),
    }

    def __init__(self, linter=None):
        super().__init__(linter)
        self._loop_level = 0
        self._loop_assignments: List[Set[str]] = []
        self._loop_names: List[List[nodes.Name]] = []
        self._loop_consts: List[List[nodes.Const]] = []
        self._ignore: List[nodes.NodeNG] = []

    @checker_utils.check_messages("loop-invariant-statement")
    def visit_for(self, node: nodes.For) -> None:
        """Visit for loop bodies."""
        self._loop_level += 1
        if isinstance(node.target, nodes.Tuple):
            self._loop_assignments.append(set(el.name for el in node.target.elts))
        elif isinstance(node.target, nodes.AssignName):
            self._loop_assignments.append({node.target.name})
        else:
            self._loop_assignments.append(set())
        self._loop_names.append([])
        self._loop_consts.append([])
        self._ignore.append(node.iter)

    @checker_utils.check_messages("loop-invariant-statement")
    def visit_while(self, node: nodes.While) -> None:
        """Visit while loop bodies."""
        self._loop_level += 1
        self._loop_names.append([])
        self._loop_consts.append([])
        self._loop_assignments.append(set())
        self._ignore.append(node.test)

    @checker_utils.check_messages("loop-invariant-statement")
    def leave_for(self, node: nodes.For) -> None:
        self._leave_loop(node)

    @checker_utils.check_messages("loop-invariant-statement")
    def leave_while(self, node: nodes.While) -> None:
        self._leave_loop(node)

    def _leave_loop(self, node: Union[nodes.For, nodes.While]) -> None:
        """Drop loop level."""
        self._loop_level -= 1
        assigned_names = self._loop_assignments.pop()
        unassigned_names = [name_node for name_node in self._loop_names.pop() if name_node.name not in assigned_names]
        used_consts = self._loop_consts.pop()
        FRAGMENT_NODE_TYPES = (nodes.FormattedValue, nodes.Attribute, nodes.Keyword, nodes.Slice, nodes.UnaryOp)
        NAME_NODES = (nodes.Name, nodes.AssignName)
        SIDE_EFFECT_NODES = (nodes.Yield, nodes.YieldFrom, nodes.Return, nodes.Raise)

        visited_nodes: Dict[nodes.NodeNG, bool] = dict()
        for name_node in [*unassigned_names, *used_consts]:
            cur_node = name_node.parent
            invariant_node = None
            while cur_node != node:
                # Walk down parent for variant components.
                is_variant = False
                if cur_node in visited_nodes:
                    is_variant = visited_nodes[cur_node]
                else:
                    if isinstance(cur_node, nodes.Call) and isinstance(cur_node.func, nodes.Name):
                        if cur_node.func in unassigned_names:
                            is_variant = True
                        elif cur_node.func.name == 'print':  # Treat print() as a side-effect
                            is_variant = True
                    elif isinstance(cur_node, SIDE_EFFECT_NODES):
                        is_variant = True
                    if not is_variant:
                        for child in get_children_recursive(cur_node):
                            if isinstance(child, NAME_NODES) and child.name in assigned_names:
                                is_variant = True
                    visited_nodes[cur_node] = is_variant
                if not is_variant:
                    invariant_node = cur_node
                    cur_node = cur_node.parent
                else:
                    break

            if invariant_node and invariant_node not in self._ignore and not isinstance(invariant_node, FRAGMENT_NODE_TYPES):
                self.add_message("loop-invariant-statement", node=invariant_node)

    def visit_assign(self, node: nodes.Assign) -> None:
        """Track assignments in loops."""
        # we don't handle multiple assignment nor slice assignment
        if not self._loop_assignments:
            return # Skip when empty

        target = node.targets[0]
        if isinstance(target, nodes.AssignName):
            self._loop_assignments[-1].add(target.name)

    def visit_augassign(self, node: nodes.AugAssign) -> None:
        """Track assignments in loops."""
        # we don't handle multiple assignment nor slice assignment
        if not self._loop_assignments:
            return # Skip when empty

        if isinstance(node.target, nodes.AssignName):
            self._loop_assignments[-1].add(node.target.name)

    @checker_utils.check_messages("loop-invariant-global-usage")
    def visit_name(self, node: nodes.Name) -> None:
        """Look for global names"""
        if self._loop_names:
            if not checker_utils.is_builtin(node.name) and node.name != "self":
                self._loop_names[-1].append(node)

        if checker_utils.is_builtin(node.name):
            return
        scope, _ = node.lookup(node.name)
        if not isinstance(scope, nodes.Module):
            return
        if node.name in scope.globals and len(scope.globals[node.name]) > 0 and isinstance(scope.globals[node.name][0], nodes.AssignName):
            if self._loop_level > 0:
                self.add_message("loop-invariant-global-usage", node=node)

    def visit_const(self, node: nodes.Const) -> None:
        if self._loop_level == 0:
            return
        if self._loop_consts:
            self._loop_consts[-1].append(node)

    def visit_call(self, node: nodes.Call) -> None:
        """Look for method calls."""
        if not isinstance(node.func, nodes.Attribute):
            return
        if not self._loop_assignments:
            return # Skip when empty
        if isinstance(node.func.expr, nodes.Name):
            self._loop_assignments[-1].add(node.func.expr.name)

    @checker_utils.check_messages('loop-try-except-usage')
    def visit_tryexcept(self, node: nodes.TryExcept) -> None:
        if self._loop_level > 0:
            self.add_message("loop-try-except-usage", node=node)

    @checker_utils.check_messages('memoryview-over-bytes')
    def visit_subscript(self, node: nodes.Subscript) -> None:
        if self._loop_level == 0:
            return
        inferred_value = safe_infer(node.value)
        if not inferred_value:
            inferred_value = local_type(node.value)
            if isinstance(inferred_value, nodes.Name) and inferred_value.name == 'bytes':
                self.add_message("memoryview-over-bytes", node=node)
        if isinstance(inferred_value, nodes.Const) and isinstance(inferred_value.value, bytes):
            self.add_message("memoryview-over-bytes", node=node)

    @checker_utils.check_messages('dotted-import-in-loop')
    def visit_attribute(self, node: nodes.Attribute) -> None:
        if self._loop_level == 0:
            return
        inferred_value = safe_infer(node.expr)
        if inferred_value and isinstance(inferred_value, nodes.Module):
            if isinstance(node.parent, nodes.Attribute):  # TODO: Go higher in the chain
                self.add_message("dotted-import-in-loop", node=node.parent, args=(node.parent.attrname,))
            self.add_message("dotted-import-in-loop", node=node, args=(node.attrname,))



import astroid
import perflint.comprehension_checker

from base import BaseCheckerTestCase


class TestComprehensionChecker(BaseCheckerTestCase):
    CHECKER_CLASS = perflint.comprehension_checker.ComprehensionChecker

    def test_simple_list_filter(self):
        test_func = astroid.extract_node(
            """
        def test(): #@
            items = [1,2,3,4]
            result = []
            for i in items:
                if i % 2:
                    result.append(i)
        """
        )

        with self.assertAddedMessage("use-list-comprehension"):
            self.walk(test_func)

    def test_complex_list_filter(self):
        test_func = astroid.extract_node(
            """
        def test(): #@
            items = [1,2,3,4]
            result = []
            for i in items:
                if i % 2:
                    result.append(i)
                elif i % 2:
                    result.append(i)
                else:
                    result.append(i)
        """
        )

        with self.assertNoMessages():
            self.walk(test_func)

    def test_simple_list_copy(self):
        test_func = astroid.extract_node(
            """
        def test(): #@
            items = [1,2,3,4]
            result = []
            for i in items:
                result.append(i)
        """
        )

        with self.assertAddedMessage("use-list-copy"):
            self.walk(test_func)

    def test_simple_dict_assignment(self):
        test_func = astroid.extract_node(
            """
        def test(): #@
            result = {}
            fruit = ["apple", "pear", "orange"]
            for idx, name in enumerate(fruit):
                result[idx] = name
        """
        )

        with self.assertAddedMessage("use-dict-comprehension"):
            self.walk(test_func)

    def test_filtered_dict_assignment(self):
        test_func = astroid.extract_node(
            """
        def test(): #@
            result = {}
            fruit = ["apple", "pear", "orange"]
            for idx, name in enumerate(fruit):
                if idx % 2:
                    result[idx] = name
        """
        )

        with self.assertAddedMessage("use-dict-comprehension"):
            self.walk(test_func)

    def test_complex_filtered_dict_assignment(self):
        test_func = astroid.extract_node(
            """
        def test(): #@
            result = {}
            fruit = ["apple", "pear", "orange"]
            for idx, name in enumerate(fruit):
                if idx % 2:
                    result[idx] = name
                elif idx % 3:
                    result[idx] = name
                else:
                    result[idx] = name
        """
        )

        with self.assertNoMessages():
            self.walk(test_func)

    def test_complex_filtered_dict_assignment2(self):
        test_func = astroid.extract_node(
            """
        def test(): #@
            result = {}
            fruit = ["apple", "pear", "orange"]
            for idx, name in enumerate(fruit):
                if idx % 2:
                    result[idx] = name
                else:
                    result[idx] = name
        """
        )

        with self.assertNoMessages():
            self.walk(test_func)



import pylint
from pylint.lint import Run as PylintRun
import sys

from perflint.for_loop_checker import ForLoopChecker, LoopInvariantChecker
from perflint.list_checker import ListChecker
from perflint.comprehension_checker import ComprehensionChecker


pylint.modify_sys_path()

rules = (
    list(ForLoopChecker.msgs.keys())
    + list(LoopInvariantChecker.msgs.keys())
    + list(ListChecker.msgs.keys())
    + list(ComprehensionChecker.msgs.keys())
)

args = []
args.append("--load-plugins=perflint")
args.append("--disable=all")
args.append("--enable={0}".format(",".join(rules)))
args.extend(sys.argv[1:])

try:
    PylintRun(args)
except KeyboardInterrupt:
    sys.exit(1)



import astroid
import perflint.list_checker

from base import BaseCheckerTestCase


class TestListMutationChecker(BaseCheckerTestCase):
    CHECKER_CLASS = perflint.list_checker.ListChecker

    def test_const_list_should_be_tuple(self):
        test_func = astroid.extract_node("""
        def test(): #@
            items = [1,2,3,4]
        """)

        with self.assertAddedMessage("use-tuple-over-list"):
            self.walk(test_func)

    def test_mutated_list_by_method(self):
        test_func = astroid.extract_node("""
        def test(): #@
            items = [1,2,3,4]
            items.clear()
        """)

        with self.assertNoMessages():
            self.walk(test_func)

    def test_mutated_list_by_index(self):
        test_func = astroid.extract_node("""
        def test(): #@
            items = [1,2,3,4]
            items[0] = 0
        """)

        with self.assertNoMessages():
            self.walk(test_func)

    def test_mutated_global_list_by_index(self):
        test_func = astroid.extract_node("""
        items = [1,2,3,4]
        def test(): #@
            items[0] = 0
        """)

        with self.assertNoMessages():
            self.walk(test_func)
    
    def test_mutated_global_list_by_method(self):
        test_func = astroid.extract_node("""
        items = [1,2,3,4]
        def test(): #@
            items.append(5)
        """)

        with self.assertNoMessages():
            self.walk(test_func)